/*
 
 Copyright (C) 2010,2011 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include "interface.h"
#include "dbio.h"
#include "delve.h"
#include "align_to_sam.h"


void unaligned_to_sam(struct read_info* ri)
{
	int i,c;
	char nuc[] = "ACGTN";
	char pline[500];
	char pseq[500];
	for(i = 0;i < ri->len;i++){
		if(ri->seq[i] == 12){
			pseq[i] = 'N';
		}else{
			pseq[i] = nuc[(int)ri->seq[i] & 3 ];
		}
	}
	pseq[ri->len]  = 0;
	
	
	if(ri->qual){
		sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
				ri->name ,//1 QNAME Query NAME of the read or the read pair 
				4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
				"*",//3 RNAME Reference sequence NAME 
				0,//4 POS 1-based leftmost POSition of clipped alignment 
				0,//5 MAPQ MAPping Quality (Phred-scaled) 
				"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
				"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
				0,//8 MPOS 1-based leftmost Mate POSition 
				0,//9 ISIZE inferred Insert SIZE 
				pseq,//10 SEQ query SEQuence on the same strand as the reference 
				ri->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
				);
	}else{
		sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
				ri->name ,//1 QNAME Query NAME of the read or the read pair 
				4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
				"*",//3 RNAME Reference sequence NAME 
				0,//4 POS 1-based leftmost POSition of clipped alignment 
				0,//5 MAPQ MAPping Quality (Phred-scaled) 
				"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
				"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
				0,//8 MPOS 1-based leftmost Mate POSition 
				0,//9 ISIZE inferred Insert SIZE 
				pseq,//10 SEQ query SEQuence on the same strand as the reference 
				"*"//11 QUAL query QUALity (ASCII-33=Phred base quality) 
				);
	}
	c = strlen(pline);
	pline[c] = 0;
	fprintf(outfile, "%s\n",pline);
}

void align_to_sam(unsigned char* aln,unsigned char* t_name,unsigned int pos, unsigned flag,float score, struct read_info* ri,unsigned char* posteriors)
{
	char nuc[] = "ACGTN";
	char cigarline[128];
	char mdline[128];
	char pline[500];
	char pseq[500];
	//unsigned char* qual = 0;
	
	//unsigned char noqual[] = "*";
	int mismatches = 0;
	int i,j,c,g,t;
	unsigned int add = 0;
	c = 0;
	int state = 0;
	mdline[0] = 0;
	c = aln[0];
	j = 0;
	t = 1;

	while((aln[c]  >> 4) == 5 ){
	//	fprintf(stderr,"%d	%d	len: %d\n",((aln[c]   >> 4)& 0xF ),(aln[c]  & 0xF),ri->len);
		c--;
	}
	for(i = c;i >= 1;i--){
	//	fprintf(stderr,"%d	%d	%d\n",((aln[i]   >> 4)& 0xF ),(aln[i]  & 0xF),t);
		if((aln[i]  >> 4) != 5 && (aln[i]  & 0xF) !=5){
			if((aln[i]  >> 4) != (aln[i]  & 0xF)){
				g = (int)strlen(mdline);
				sprintf(mdline+g,"%d%c",j,nuc[(aln[i]  & 0xF)]);
				//fprintf(stderr,"%d	%c\n",j,nuc[(aln[i]  & 0xF)]);
				j = 0;
			}else{
				j++;
			}
			
			if(t == ri->len){
				i = -1;
				break;
			}
			t++;
			 state = 0;
		}else if((aln[i]  >> 4) != 5 && (aln[i]  & 0xF) == 5){
			
				
			
			if(t == ri->len){
				i = -1;
				break;
			}
			t++;
			
			//fprintf(stderr,"%c	\n",nuc[(aln[i]  & 0xF)]);
			//j = 0;
			state = 0;
		}else if((aln[i]  >> 4) == 5  && (aln[i]  & 0xF) != 5){
			if(state == 0){
				g = (int)strlen(mdline);
				sprintf(mdline+g,"%d^",j);
				//fprintf(stderr,"%d	\n",j);
				state = 1;
			}
			g = (int)strlen(mdline);
			sprintf(mdline+g,"%c",nuc[(aln[i]  & 0xF )]);
			j = 0;
			//state = 0;
		}
	}
	
	if(j){
		g = (int)strlen(mdline);
		sprintf(mdline+g,"%d",j);
	//	fprintf(stderr,"%d	\n",j);
	}
	
	g = (int)strlen(mdline);
	if(!isdigit((int)  mdline[g-1])){
		sprintf(mdline+g,"%d",0);
	}
	
	//fprintf(stderr,"LINE:%s\n\n",mdline);
	
	
	i =  aln[0];
	t = 0;
	cigarline[0] = 0;
	while((aln[i]  >> 4) == 5){
		i--;
		add++;
	}
	
	while(i >= 1){
		j = 0;
		if((aln[i]  >> 4) != 5 && (aln[i]  & 0xF) != 5){
			while ( (aln[i]  >> 4) != 5 && (aln[i]  & 0xF) != 5 && i >= 1) {
				if((aln[i]  >> 4) != (aln[i]  & 0xF) ){
					mismatches++;
				}
				i--;
				j++;
				t++;
				
			}
			c = (int)strlen(cigarline);
			sprintf(cigarline+c,"%dM",j);
			if(t == ri->len){
				break;
			}
			
		}else if((aln[i]  >> 4) != 5 && (aln[i]  & 0xF) == 5 ){
			while ((aln[i]  >> 4) != 5 && (aln[i]  & 0xF) == 5&& i >= 1) {
				i--;
				j++;
				t++;
				mismatches++;
			}
			c = (int)strlen(cigarline);
			sprintf(cigarline+c,"%dI",j);
			if(t == ri->len){
				break;
			}
		}else if((aln[i]  >> 4) == 5 && (aln[i]  & 0xF) != 5 ){
			while ((aln[i]  >> 4) == 5 && (aln[i]  & 0xF) != 5&& i >= 1) {
				i--;
				j++;
				mismatches++;
			}
			c = (int)strlen(cigarline);
			sprintf(cigarline+c,"%dD",j);
		}
	//	fprintf(stderr,"%d	%s\n",i,cigarline);
	}
	//fprintf(stderr,"CIGAR ::: %s\n",cigarline);
	
	for(i = 0;i < ri->len;i++){
		if(ri->seq[i] == 12){
			pseq[i] = 'N';
		}else{
			pseq[i] = nuc[(int)ri->seq[i] & 3 ];
		}
	}
	pseq[ri->len]  = 0;
	
	/*if(ri->qual){
		if(strand){
			ri->qual = reverse_without_complement(ri->qual, ri->len);
		}
		qual = ri->qual;
		
		
	}else{
		qual = noqual;
	}*/
		
	if(1.0f - score < 0.0001f){
		score = 0.0001f;
	}else{
		score = 1.0f - score;
	}
	
	
	//fprintf(stderr,"Phred calculation:     = -10 * log10 (%f) = -10 * %f \n",score, log10(score));
	
	
	
	if(flag & 0x100){
		sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
				ri->name ,//1 QNAME Query NAME of the read or the read pair 
				flag,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
				t_name,//3 RNAME Reference sequence NAME 
				pos+add +1 ,//4 POS 1-based leftmost POSition of clipped alignment 
				(int)((-10.0f * log10(score))+0.5f),//5 MAPQ MAPping Quality (Phred-scaled) 
				cigarline,//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
				"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
				0,//8 MPOS 1-based leftmost Mate POSition 
				0,//9 ISIZE inferred Insert SIZE 
				"*" ,//10 SEQ query SEQuence on the same strand as the reference 
				"*"//11 QUAL query QUALity (ASCII-33=Phred base quality) 
				);
	}else{
		sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
				ri->name ,//1 QNAME Query NAME of the read or the read pair 
				flag,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
				t_name,//3 RNAME Reference sequence NAME 
				pos+add +1 ,//4 POS 1-based leftmost POSition of clipped alignment 
				(int)((-10.0f * log10(score))+0.5f),//5 MAPQ MAPping Quality (Phred-scaled) 
				cigarline,//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
				"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
				0,//8 MPOS 1-based leftmost Mate POSition 
				0,//9 ISIZE inferred Insert SIZE 
				pseq ,//10 SEQ query SEQuence on the same strand as the reference 
				ri->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
				);
	}
	
	i = strlen(pline);
	pline[i] = '\t';
	i++;
	sprintf(pline+i,"NM:i:%d",  mismatches);
	
	i = strlen(pline);
	pline[i] = '\t';
	i++;
	sprintf(pline+i,"MD:Z:%s",  mdline);
	
	if(posteriors){
		i = strlen(pline);
		pline[i] = '\t';
		i++;
		sprintf(pline+i,"XP:Z:%s",posteriors);
	}
	fprintf(outfile, "%s\n",pline);
	//if(ri->qual){
	//	if(strand){
	//		ri->qual = reverse_without_complement(ri->qual, ri->len);
	//	}
	//}		
	////fprintf(stderr, "%s\n",pline);
	
}

